# file generated by setuptools_scm
# don't change, don't track in version control
from __future__ import annotations

__version__ = version = '2.5.0'  # type: str
__version_tuple__ = version_tuple = (2, 5, 0)  # type: tuple[int | str, ...]
